@@ -6,17 +6,21 @@ IPDR - IPDR Data Suite
 
 =head1 VERSION
 
-Version 0.29_01
+Version 0.41
 
 =cut
 
-our $VERSION = '0.29_01';
+our $VERSION = '0.41';
 
 =head1 SYNOPSIS
 
 This is a IPDR collection suite for currently Cisco and generic IPDR compliant
 servers.
 
+To use the XDR module to extract XDR files see the man page
+
+IPDR::Process::XDR
+
 To use the Cisco implementation of IPDR (SAMIS) see the man page
 
 IPDR::Collection::Cisco
@@ -78,7 +82,7 @@ L<http://search.cpan.org/dist/IPDR>
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2009 Andrew S. Kennedy, all rights reserved.
+Copyright 2011 Andrew S. Kennedy, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
